/**
 * Copyright (C) 2014 Deepin Technology Co., Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 **/

package dock

import (
	"errors"
	"path/filepath"
	"strings"

	"gir/gio-2.0"
	. "pkg.deepin.io/lib/gettext"
	"pkg.deepin.io/lib/utils"
)

type NormalApp struct {
	Id        string
	DesktopID string
	Icon      string
	Name      string
	Menu      string

	changedCB func()

	path     string
	coreMenu *Menu
	dockItem *MenuItem
}

func NewNormalApp(desktopID string) *NormalApp {
	app := &NormalApp{Id: normalizeAppID(trimDesktop(desktopID)), DesktopID: desktopID}
	var core *AppInfo
	if strings.ContainsRune(desktopID, filepath.Separator) {
		core = NewAppInfoFromFile(desktopID)
		app.Id = filepath.Base(app.Id)
	} else {
		core = NewAppInfo(desktopID)
		if core == nil {
			newId := guess_desktop_id(app.Id)
			logger.Debugf("guess desktop: %q", newId)
			if newId != "" {
				core = NewAppInfo(newId)
				app.DesktopID = newId
			}
		}
	}
	logger.Info("NewNormalApp:", app.Id, "for desktop", desktopID)
	if core == nil {
		return nil
	}
	defer core.Destroy()
	app.path = core.GetFilename()
	app.Icon = core.GetIcon()
	logger.Debug(app.Id, "::app icon:", app.Icon)
	app.Name = core.GetDisplayName()
	logger.Debug("Name", app.Name)
	app.buildMenu(core)
	return app
}

func (app *NormalApp) createDesktopAppInfo() *AppInfo {
	core := NewAppInfo(app.DesktopID)

	if core != nil {
		return core
	}

	if newId := guess_desktop_id(app.Id); newId != "" {
		core = NewAppInfo(newId)
		if core != nil {
			return core
		}
	}

	return NewAppInfoFromFile(app.path)
}

func (app *NormalApp) buildMenu(core *AppInfo) {
	app.coreMenu = NewMenu()
	app.coreMenu.AppendItem(NewMenuItem(Tr("_Run"), func(timestamp uint32) {
		core := app.createDesktopAppInfo()
		if core == nil {
			logger.Warning("CreateDesktopAppInfo failed")
			return
		}
		defer core.Destroy()
		_, err := core.Launch(make([]*gio.File, 0), gio.GetGdkAppLaunchContext().SetTimestamp(timestamp))
		if err != nil {
			logger.Warning("Launch App Failed: ", err)
		}
	}, true))
	app.coreMenu.AddSeparator()
	for _, actionName := range core.ListActions() {
		name := actionName //NOTE: don't directly use 'actionName' with closure in an forloop
		app.coreMenu.AppendItem(NewMenuItem(
			core.GetActionName(actionName),
			func(timestamp uint32) {
				core := app.createDesktopAppInfo()
				if core == nil {
					logger.Warning("start action", name,
						"failed")
					return
				}
				defer core.Destroy()
				core.LaunchAction(name, gio.GetGdkAppLaunchContext().SetTimestamp(timestamp))
			},
			true,
		))
	}
	app.coreMenu.AddSeparator()
	dockItem := NewMenuItem(
		Tr("_Undock"),
		func(timestamp uint32) {
			dockManager.entryManager.dockedAppManager.Undock(app.Id)
		},
		true,
	)
	app.coreMenu.AppendItem(dockItem)

	app.Menu = app.coreMenu.GenerateJSON()
}

func (app *NormalApp) HandleMenuItem(id string, timestamp uint32) {
	if app.coreMenu != nil {
		app.coreMenu.HandleAction(id, timestamp)
	}
}

func NewNormalAppFromFilename(name string) *NormalApp {
	app := &NormalApp{}
	appInfo := NewAppInfoFromFile(name)
	if appInfo == nil {
		return nil
	}
	defer appInfo.Destroy()
	app.Id = appInfo.GetId()
	app.DesktopID = app.Id + ".desktop"
	app.path = appInfo.GetFilename()
	app.Icon = appInfo.GetIcon()
	app.Name = appInfo.GetDisplayName()
	app.buildMenu(appInfo)
	return app
}

func (app *NormalApp) Activate(x, y int32, timestamp uint32) error {
	// FIXME:
	// the launch will be successful even if the desktop file is not
	// existed.
	if !utils.IsFileExist(app.path) {
		return errors.New("invalid")
	}

	core := app.createDesktopAppInfo()
	if core == nil {
		return errors.New("create desktop app info failed")
	}
	defer core.Destroy()
	logger.Debug("launch desktop file:", core.GetFilename())
	_, err := core.Launch(nil, gio.GetGdkAppLaunchContext().SetTimestamp(timestamp))
	if err != nil {
		logger.Warning("launch", app.Id, "failed:", err)
	}
	return err
}

func (app *NormalApp) setChangedCB(cb func()) {
	app.changedCB = cb
}

func (app *NormalApp) notifyChanged() {
	if app.changedCB != nil {
		app.changedCB()
	}
}

func (app *NormalApp) HandleDragDrop(path string, timestamp uint32) {
	logger.Debugf("handle drag drop path: %q", path)
	ai := app.createDesktopAppInfo()
	if ai != nil {
		defer ai.Destroy()
		paths := []string{path}
		_, err := ai.LaunchUris(paths, gio.GetGdkAppLaunchContext().SetTimestamp(timestamp))
		if err != nil {
			logger.Warning("LaunchUris failed:", err)
		}
	}
}
