/**
 * Copyright (C) 2014 Deepin Technology Co., Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 **/

package sessionwatcher

import (
	"dbus/org/freedesktop/login1"
	"pkg.deepin.io/lib/dbus"
	"sync"
)

const (
	login1Dest = "org.freedesktop.login1"
	login1Path = "/org/freedesktop/login1"
)

type Manager struct {
	loginManager  *login1.Manager
	sessionLocker sync.Mutex
	sessions      map[string]*login1.Session
}

func newManager() (*Manager, error) {
	loginObj, err := login1.NewManager(login1Dest, login1Path)
	if err != nil {
		logger.Warning("New login1 manager failed:", err)
		return nil, err
	}

	return &Manager{
		loginManager: loginObj,
		sessions:     make(map[string]*login1.Session),
	}, nil
}

func (m *Manager) destroy() {
	if m.sessions != nil {
		m.destroySessions()
	}
}

func (m *Manager) initUserSessions() {
	list, err := m.loginManager.ListSessions()
	if err != nil {
		logger.Warning("List sessions failed:", err)
		return
	}

	for _, v := range list {
		// v info: (id, uid, username, seat id, session path)
		if len(v) != 5 {
			logger.Warning("Invalid session info:", v)
			continue
		}

		id, ok := v[0].(string)
		if !ok {
			continue
		}

		p, ok := v[4].(dbus.ObjectPath)
		if !ok {
			continue
		}

		m.addSession(id, p)
	}

	m.loginManager.ConnectSessionNew(func(id string, path dbus.ObjectPath) {
		logger.Debug("Session added:", id, path)
		m.addSession(id, path)
	})

	m.loginManager.ConnectSessionRemoved(func(id string, path dbus.ObjectPath) {
		logger.Debug("Session removed:", id, path)
		m.deleteSession(id, path)
	})
}

func (m *Manager) destroySessions() {
	m.sessionLocker.Lock()
	for _, s := range m.sessions {
		login1.DestroySession(s)
		s = nil
	}
	m.sessions = nil
	m.sessionLocker.Unlock()
}

func (m *Manager) addSession(id string, path dbus.ObjectPath) {
	uid, session := newLoginSession(path)
	if session == nil {
		return
	}

	logger.Debug("Add session:", id, path, uid)
	if !isCurrentUser(uid) {
		logger.Debug("Not the current user session:", id, path, uid)
		return
	}

	m.sessionLocker.Lock()
	m.sessions[id] = session
	m.sessionLocker.Unlock()

	session.Active.ConnectChanged(func() {
		if session == nil {
			return
		}
		m.handleSessionChanged()
	})
	m.handleSessionChanged()
}

func (m *Manager) deleteSession(id string, path dbus.ObjectPath) {
	m.sessionLocker.Lock()
	session, ok := m.sessions[id]
	if !ok {
		m.sessionLocker.Unlock()
		return
	}

	logger.Debug("Delete session:", id, path)
	login1.DestroySession(session)
	session = nil
	delete(m.sessions, id)
	m.sessionLocker.Unlock()
	m.handleSessionChanged()
}

func (m *Manager) handleSessionChanged() {
	if len(m.sessions) == 0 {
		return
	}

	if m.isActive() {
		suspendPulseSinks(0)
		suspendPulseSources(0)
	} else {
		suspendPulseSinks(1)
		suspendPulseSources(1)
	}
}

func (m *Manager) isActive() bool {
	var active bool = false
	m.sessionLocker.Lock()
	for _, s := range m.sessions {
		logger.Debug("[isActive] info:", s.Path, s.Active.Get())
		if s.Active.Get() {
			active = true
			break
		}
	}
	m.sessionLocker.Unlock()

	logger.Debug("Session state:", curUid, active)
	return active
}
